#ifndef PBBAM_ACCURACY_H
#define PBBAM_ACCURACY_H

#include <pbbam/Config.h>

#include <pbcopper/data/Accuracy.h>

namespace PacBio {
namespace BAM {

using Accuracy PBBAM_DEPRECATED = Data::Accuracy;

}  // namespace BAM
}  // namespace PacBio

#endif  // PBBAM_ACCURACY_H
