#include "PbbamInternalConfig.h"

#include <pbbam/IFastaWriter.h>

namespace PacBio {
namespace BAM {

IFastaWriter::IFastaWriter() = default;

IFastaWriter::~IFastaWriter() = default;

}  // namespace BAM
}  // namespace PacBio
