#ifndef PBBAM_ORIENTATION_H
#define PBBAM_ORIENTATION_H

#include <pbbam/Config.h>

#include <pbcopper/data/Orientation.h>

namespace PacBio {
namespace BAM {

using Orientation PBBAM_DEPRECATED = PacBio::Data::Orientation;

}  // namespace BAM
}  // namespace PacBio

#endif  // PBBAM_ORIENTATION_H
