#ifndef PBBAM_VIRTUALPOLYMERASEBAMRECORD_H
#define PBBAM_VIRTUALPOLYMERASEBAMRECORD_H

#include <pbbam/Config.h>

#include <pbbam/virtual/VirtualZmwBamRecord.h>

namespace PacBio {
namespace BAM {

/// \deprecated Use VirtualZmwBamRecord instead.
using VirtualPolymeraseBamRecord = VirtualZmwBamRecord;

}  // namespace BAM
}  // namespace PacBio

#endif  // PBBAM_VIRTUALPOLYMERASEBAMRECORD_H
