#ifndef PBBAM_FRAMES_H
#define PBBAM_FRAMES_H

#include <pbbam/Config.h>

#include <pbcopper/data/Frames.h>

namespace PacBio {
namespace BAM {

using Frames PBBAM_DEPRECATED = PacBio::Data::Frames;

}  // namespace BAM
}  // namespace PacBio

#endif  // PBBAM_FRAMES_H
