#include "PbbamInternalConfig.h"

#include <pbbam/IRecordWriter.h>

namespace PacBio {
namespace BAM {

IRecordWriter::IRecordWriter() = default;

IRecordWriter::~IRecordWriter() = default;

}  // namespace BAM
}  // namespace PacBio
